--
--  Keep Drive Spinning
--
--  Created by Jon Stovell on 2015-03-02.
--  Copyright (c) 2015 Jon Stovell. All rights reserved.
--


-- INSTRUCTIONS
--
-- To build this from source for yourself, open this file
-- in Script Editor and use File > Export... to save it
-- as an application. Then run the application.


script AppDelegate
	property parent : class "NSObject"
	
	-- IBOutlets
	property theWindow : missing value
	
	-- Global variables are handy sometimes
	global launchagent_path
	global my_id
	-- The interval used to be configurable by the user on a per-drive basis
	-- It is retained as a variable in case I bring that possibility back in the future
	global interval
	-- This exists merely to avoid showing unnecessary dialogs to the user
	global entered_admin_password
	
	on applicationWillFinishLaunching:aNotification
		set my_id to current application's id
		set theArgs to current application's NSProcessInfo's processInfo()'s arguments()
		mainscript(theArgs)
		quit
	end applicationWillFinishLaunching:
	
	-- Decides what to do
	on mainscript(theArgs)
		set launchagent_path to POSIX path of (path to library folder from user domain) & "LaunchAgents/info.stovell.KeepDriveSpinning_agent.plist"
		set entered_admin_password to false
		set interval to missing value
		
		-- This needs to be correct for the launch agent
		try
			get my_id
		on error
			display alert "You must save this script as an application before running it." as critical
			return
		end try
		
		-- Filter out any argument that is obviously not a POSIX path
		set drivePaths to {}
		repeat with i from 2 to count theArgs
			set thisArg to (item i of theArgs) as text
			if first character of thisArg is "/" then set end of drivePaths to thisArg
		end repeat
		
		if drivePaths is {} then
			showUI()
		else
			set errors_encountered to {}
			repeat with drivePath in drivePaths
				set pokeResult to pokeDrive(drivePath as text)
				if pokeResult is not true then set end of errors_encountered to {p:drivePath, m:pokeResult}
			end repeat
			
			-- Uh-oh
			if (count errors_encountered) > 0 then dealWithErrors(errors_encountered)
		end if
	end mainscript
	
	-- Updates the hidden file on a drive
	on pokeDrive(drivePath)
		try
			-- First cd to the directory, just in case the path is messy.
			-- Then test that our hidden file is writable and is not a symbolic link.
			do shell script "if [[ -d " & quoted form of drivePath & " ]]; then cd " & quoted form of drivePath & "; if [[ -w .keepThisDriveSpinning && ! -L .keepThisDriveSpinning ]]; then date '+%F %T %Z' > .keepThisDriveSpinning; fi; fi"
			return true
		on error errMsg number errNum
			return errMsg
		end try
	end pokeDrive
	
	on dealWithErrors(errors_encountered)
		if (count errors_encountered) = 1 then
			set dialog_text to "Uh-oh! Keep Drive Spinning ran into a problem while trying to keep this drive awake:" & return & p of item 1 of errors_encountered & return & return & "The error was:" & return & m of item 1 of errors_encountered
		else
			set dialog_text to "Uh-oh! Keep Drive Spinning ran into problems while trying to keep these drives awake:"
			repeat with i from 1 to count errors_encountered
				set dialog_text to return & return & p of item i of errors_encountered & return & m of item i of errors_encountered
			end repeat
		end if
		
		display dialog dialog_text buttons {"Change settings", "Ignore", "Try to fix it"} default button "Try to fix it" with icon caution with title "Keep Drive Spinning"
		
		if button returned of result is "Try to fix it" then
			repeat with i from 1 to count errors_encountered
				createHiddenFile(p of item i of errors_encountered)
			end repeat
			
		else if button returned of result is "Ignore" then
			set current_vols to getCurrentVols(launchagent_path)
			set working_vols to {}
			set error_vols to {}
			
			repeat with e in errors_encountered
				set end of error_vols to p of e
			end repeat
			
			repeat with i from 1 to count current_vols
				set c to item i of current_vols
				if c is not in error_vols then set end of working_vols to c
			end repeat
			
			updateLaunchAgent(working_vols)
			
		else
			showUI()
		end if
	end dealWithErrors
	
	-- Shows the user interface for configuration
	on showUI()
		-- This shows the app icon in the Dock
		try
			set current application's NSApp's activationPolicy to 0
		end try
		
		-- Make sure the upgrade procedure has been performed before proceeding
		upgrade()
		
		set done to false
		repeat until done
			set main_action to button returned of (display dialog "Keep Drive Spinning prevents drives from going into low-power mode." & return & return & "To get started, click the Choose Drives button." & return & return & "Click the Done button to save your settings and close this window. Your drives will be kept awake using a background process." & return & return & "You can run Keep Drive Spinning again at any time to change your settings." buttons {"Done", "More...", "Choose Drives"} default button "Choose Drives" with icon note with title "Keep Drive Spinning")
			
			if main_action is "Choose Drives" then
				configure()
			else if main_action is "More..." then
				more_actions()
			else
				set done to true
			end if
		end repeat
	end showUI
	
	-- Prompts the user to select the drives to keep awake
	on configure()
		try
			-- sleepless will hold the updated list of volumes to keep awake
			set sleepless to {}
			set sleepless_names to {}
			
			-- Get a list of names and paths for all eligible mounted drives
			set vols_info to {}
			set vols to {}
			tell application "Finder"
				set mounted_disks to every disk whose free space > 0 and (ejectable is true or local volume is false) and URL does not contain "file:///Volumes/com.apple.TimeMachine."
				repeat with i from 1 to count mounted_disks
					set vol_path to POSIX path of (item i of mounted_disks as alias)
					if vol_path is not "/" and last character of vol_path is "/" then set vol_path to (characters 1 thru -2 of vol_path) as text
					set end of vols_info to {p:vol_path, n:name of item i of mounted_disks}
					set end of vols to vol_path
				end repeat
			end tell
			copy vols to mounted_vols

			-- This adds entries for known drives that are not currently mounted
			set current_vols to getCurrentVols(launchagent_path, false)
			repeat with i from 1 to count current_vols
				set vol to (item i of current_vols) as text
				if vols does not contain vol and vol is not "" then
					set volname to last item of delimitText("/", vol)
					set end of vols_info to {p:vol, n:volname}
					set end of vols to vol
				end if
			end repeat
			
			if (count vols) = 0 then error "No eligible drives were found. Keep Drive Spinning only works with external drives and network shares that are writable."
			
			-- Now make a user-friendly list of just the drive names to choose from
			set current_volnames to getCurrentVols(launchagent_path, true)
			set volnames to {}
			repeat with vol_info in vols_info
				set end of volnames to n of vol_info
			end repeat
			
			-- Ask the user
			set selected to choose from list volnames default items current_volnames with prompt "Select the drives to keep awake." & return & return & "Press the Command key to select or deselect multiple items." with title "Keep Drive Spinning" with multiple selections allowed and empty selection allowed
			
			-- What should we do now?
			if selected is false then
				error number -128
			else if selected is {} then
				letAllSleep()
			else
				repeat with i from 1 to count selected
					set sel to item i of selected
					
					-- Get the path of the selected drive
					repeat with j from 1 to count vols_info
						if n of item j of vols_info is sel then
							set sel_path to p of item j of vols_info
							exit repeat
						end if
					end repeat
					
					-- Skip unmounted drives
					if mounted_vols contains sel_path then
						-- Set the name and path to our hidden file on the drive
						createHiddenFile(sel_path, sel)
					end if
					
					set end of sleepless to sel_path
					set end of sleepless_names to sel
				end repeat
				
				updateLaunchAgent(sleepless)
				
				display dialog "The following drives will be kept awake:" & return & return & concatenateList(return, sleepless_names) buttons {"OK"} default button "OK" with icon note with title "Keep Drive Spinning"
			end if
		on error errMsg number errNum
			try
				if errNum is not -128 then
					beep
					display dialog errMsg buttons {"Cancel"} default button "Cancel" with icon stop with title "Keep Drive Spinning"
				end if
			end try
		end try
	end configure
	
	on more_actions()
		set available_actions to {"Let all drives sleep", "Add a custom location", "Customize update schedule", "Run diagnostic"}
		set selected_action to (choose from list available_actions with prompt "What would you like to do?" with title "Keep Drive Spinning")
		if selected_action is false then return
		set selected_action to item 1 of selected_action
		
		if selected_action is "Let all drives sleep" then
			letAllSleep()
		else if selected_action is "Add a custom location" then
			set dir to POSIX path of (choose folder with prompt "Where should Keep Drive Spinning put its hidden file?")
			if dir is not "/" and last character of dir is "/" then set dir to (characters 1 thru -2 of dir) as text
			createHiddenFile(dir, dir)
			set vols to getCurrentVols(launchagent_path, false) & dir
			updateLaunchAgent(vols)
			
			set vols to getCurrentVols(launchagent_path, true)
			display dialog "The following drives will be kept awake:" & return & return & concatenateList(return, vols) buttons {"OK"} default button "OK" with icon note with title "Keep Drive Spinning"
		else if selected_action is "Customize update schedule" then
			-- Ask for a time interval
			repeat
				set interval to text returned of (display dialog "How often (in seconds) should Keep Drive Spinning poke the drives to keep them awake?" default answer "60" buttons {"OK"} default button "OK" with icon note with title "Keep Drive Spinning")
				try
					set interval to interval as number
					if interval is less than 11 then
						display dialog "The operating system will not permit the job to run this frequently. The default value of 60 seconds is recommended." buttons {"OK"} default button "OK" with icon note with title "Keep Drive Spinning"
						error
					else if interval is less than 60 then
						display dialog "An interval of less than 60 seconds is not recommended. It won't work any better to poke the drives that frequently, and it may negatively affect your computer's performance." & return & return & "Are you sure you want to use an interval of " & interval & " seconds?" buttons {"Yes", "No"} default button "No" with icon note with title "Keep Drive Spinning"
						if button returned of result is "No" then error
					end if
					exit repeat
				end try
			end repeat
			set vols to getCurrentVols(launchagent_path, false)
			updateLaunchAgent(vols)
		else if selected_action is "Run diagnostic" then
			runDiagnostic()
			activate
		end if
	end more_actions
	
	-- Creates/fixes a .keepThisDriveSpinning file
	on createHiddenFile(drivePath, driveName)
		set hiddenfile to drivePath & "/.keepThisDriveSpinning"
		
		-- Just in case...
		repeat while hiddenfile contains "//"
			set hiddenfile to findReplace("//", "/", hiddenfile)
		end repeat
		
		-- If the hidden file already exists, make sure it isn't locked
		try
			do shell script "if [[ -f " & quoted form of (hiddenfile) & " ]]; then chflags -f nouchg " & quoted form of (hiddenfile) & "; fi"
		on error
			if entered_admin_password is false then display dialog hiddenfile & " must be unlocked for Keep Drive Spinning to work." & return & return & "Retry with an administrator password?" buttons {"Cancel", "Retry"} with icon caution with title "Keep Drive Spinning"
			do shell script "if [[ -f " & quoted form of (hiddenfile) & " ]]; then chflags -f nouchg " & quoted form of (hiddenfile) & "; fi" with administrator privileges
			set entered_admin_password to true
		end try
		
		-- Create the hidden file on the drive
		try
			do shell script "date '+%F %T %Z' > " & quoted form of hiddenfile
		on error errMsg number errNum
			if errMsg contains "Permission denied" then
				if entered_admin_password is false then
					try
						get (POSIX file (hiddenfile)) as alias
						set dialog_message to "You do not have permission to update the necessary file in " & drivePath & ". "
					on error
						set dialog_message to "You do not have permission to create the necessary file in " & drivePath & ". "
					end try
					set retry to (button returned of (display dialog dialog_message & return & return & "Retry with an administrator password?" buttons {"Cancel", "Retry"} with icon caution with title "Keep Drive Spinning") is "Retry")
				else
					set retry to true
				end if
				if retry then
					try
						do shell script "date '+%F %T %Z' > " & quoted form of hiddenfile with administrator privileges
						set entered_admin_password to true
					on error errMsg number errNum
						error "Could not create the necessary file in " & drivePath & ". The following error occurred:" & return & errMsg & return & return & "This is not an error in Keep Drive Spinning itself. Please correct the issue and then try again." number errNum
					end try
				end if
			else if errMsg contains "Read-only file system" then
				error driveName & " is a read-only file system. Drives are kept awake by updating a hidden file on the drive once per minute. Files can't be updated on read-only drives, so " & driveName & " can't be forced to stay awake." number errNum
			else
				error "Could not create the necessary file in " & drivePath & ". The following error occurred:" & return & errMsg & return & return & "This is not an error in Keep Drive Spinning itself. Please correct the issue and then try again." number errNum
			end if
		end try
		
		-- Other users might also want to keep this drive awake, so let them
		try
			do shell script "chmod 666 " & quoted form of hiddenfile
		on error
			if entered_admin_password is false then display dialog "Could not set the correct access permissions on " & hiddenfile & "." & return & return & "Retry with an administrator password?" buttons {"Cancel", "Retry"} with icon caution with title "Keep Drive Spinning"
			do shell script "chmod 666 " & quoted form of hiddenfile with administrator privileges
			set entered_admin_password to true
		end try
	end createHiddenFile
	
	-- Gets the list of volumes that the launch agent tries to keep awake
	on getCurrentVols(agent_path, name_only)
		set vols to {}
		try
			set agent_volumes to (do shell script "xmllint --xpath " & quoted form of "//array[preceding-sibling::key/text()='ProgramArguments' and string[1]/text()='/usr/bin/open']/string[position()>=5]" & space & quoted form of agent_path)
			set agent_volumes to findReplace({"<string>", "<string/>"}, "", agent_volumes)
			set agent_volumes to delimitText("</string>", agent_volumes)
			repeat with i from 1 to count agent_volumes
				set vol to item i of agent_volumes
				if vol is not "" then
					if name_only is true then
						try
							tell application "Finder" to set end of vols to name of ((POSIX file vol) as alias)
						on error
							set end of vols to last item of delimitText("/", vol)
						end try
					else
						set end of vols to (vol as text)
					end if
				end if
			end repeat
		end try
		return vols
	end getCurrentVols
	
	-- Writes a new version of the launch agent to disk and loads it
	on updateLaunchAgent(vols)
		if interval is missing value then
			try
				set interval to do shell script "defaults read " & quoted form of launchagent_path & " StartInterval"
			on error
				set interval to 60
			end try
		end if
		
		set vols to uniqueList(vols)
		
		repeat with i from 1 to count vols
			set vol_xmlsafe to item i of vols
			set vol_xmlsafe to my findReplace("&", "&amp;", vol_xmlsafe)
			set vol_xmlsafe to my findReplace("<", "&lt;", vol_xmlsafe)
			set vol_xmlsafe to my findReplace(">", "&gt;", vol_xmlsafe)
			set item i of vols to vol_xmlsafe
		end repeat
		
		try
			set plisttext to concatenateList(linefeed, {"<?xml version=\"1.0\" encoding=\"UTF-8\"?>", 
				"<!DOCTYPE plist PUBLIC \"-//Apple Computer//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">", 
				"<plist version=\"1.0\">", 
				"<dict>", 
				"  <key>Label</key>", 
				"  <string>info.stovell.KeepDriveSpinning_agent</string>", 
				"  <key>ProgramArguments</key>", 
				"  <array>", 
				"    <string>/usr/bin/open</string>", 
				"    <string>-b</string>", 
				"    <string>" & my_id & "</string>", 
				"    <string>--args</string>"})
			
			repeat with vol in vols
				set plisttext to plisttext & linefeed & "    <string>" & vol & "</string>"
			end repeat
			
			set plisttext to plisttext & concatenateList(linefeed, {"", 
				"  </array>", 
				"  <key>RunAtLoad</key>", 
				"  <true/>", 
				"  <key>StartInterval</key>", 
				"  <integer>" & interval & "</integer>", 
				"</dict>", 
				"</plist>"})
			
			-- Create and load the launch agent
			do shell script "mkdir -p \"$(dirname " & quoted form of launchagent_path & ")\"; launchctl unload -w " & quoted form of launchagent_path & " 2>/dev/null; echo " & quoted form of plisttext & " > " & quoted form of launchagent_path & "; launchctl load -w " & quoted form of launchagent_path
			
			-- Double check that it loaded properly
			set agentloaded to false
			repeat 4 times
				if (do shell script "launchctl list") does not contain "info.stovell.KeepDriveSpinning_agent" & return then
					delay 0.5
				else
					set agentloaded to true
					exit repeat
				end if
			end repeat
			if not agentloaded then
				error "Could not load the launch agent. Make sure that the correct permissions are set on " & POSIX path of (((path to library folder from user domain) as text) & "LaunchAgents")
			end if
		on error errMsg number errNum
			error "Could not create a valid launch agent." & return & errMsg number errNum
		end try
	end updateLaunchAgent
	
	on letAllSleep()
		do shell script "launchctl unload -w " & quoted form of launchagent_path & "; rm -f " & quoted form of launchagent_path
		display dialog "All drives are now allowed to sleep." buttons {"OK"} default button "OK" with icon note with title "Keep Drive Spinning"
		set interval to missing value
	end letAllSleep
	
	-- Finds legacy launch agents and migrates them to the latest format
	on upgrade()
		set vols to {}
		set obsoleteAgents to every paragraph of (do shell script "find ~/Library/LaunchAgents -name 'net.jonstovell.keep*Spinning.plist' -or -name 'info.stovell.KeepDriveSpinning.*.plist'")
		if (count obsoleteAgents) = 0 then return
		
		repeat with thisAgent in obsoleteAgents
			-- First try 2.0 format
			try
				set vol to (do shell script "xmllint --xpath " & quoted form of "//array[preceding-sibling::key/text()='ProgramArguments' and string[1]/text()='/bin/bash']/string[3]/text()" & space & quoted form of thisAgent)
				set vol to second item of delimitText({"[[ -e ", "[[ -w ", " ]]"}, vol)
				set vol to (do shell script "dirname " & vol)
				set end of vols to (vol as text)
			on error
				-- Next try 3.0 format, just in case...
				set vols to vols & getCurrentVols(thisAgent, false)
			end try
		end repeat
		
		-- Create the new launch agent using our list of volumes
		updateLaunchAgent(vols)
		
		-- Now unload and delete the obsolete agents
		repeat with thisAgent in obsoleteAgents
			do shell script "launchctl unload -w " & quoted form of thisAgent & "; rm -f " & quoted form of thisAgent
		end repeat
	end upgrade
	
	on runDiagnostic()
		try
			set current application's NSApp's activationPolicy to 0
		end try
		
		display dialog "Keep Drive Spinning will now conduct a series of tests to determine whether it is working correctlyand if not, why not." & return & return & "The diagnostic report will be saved in a file on your desktop and opened for you." with icon note with title "Keep Drive Spinning"
		
		-- Get a list of names and paths for all attached drives
		set vols to {}
		tell application "Finder"
			set mounted_disks to every disk whose free space > 0 and (ejectable is true or local volume is false) and URL does not contain "file:///Volumes/com.apple.TimeMachine."
			repeat with i from 1 to count mounted_disks
				set vol_path to POSIX path of ((item i of mounted_disks) as alias)
				if vol_path is not "/" and last character of vol_path is "/" then set vol_path to (characters 1 thru -2 of vol_path) as text
				set end of vols to vol_path
			end repeat
		end tell
		
		-- Determine exactly which volumes to inspect
		set vols to choose from list vols with prompt "Select the drives to test." & return & return & "Press the Command key to select multiple items." with title "Keep Drive Spinning" with multiple selections allowed
		if vols is false then error number -128
		
		-- Gather some basic information
		set os_version to system version of (system info)
		set currentdate to do shell script "date '+%F %T %Z'"
		set currentdate_seconds to do shell script "date -j -f '%F %T %Z' '+%s' " & quoted form of currentdate
		
		set report to {}
		repeat with i from 1 to count vols
			try
				set thisReport to {}
				set vol_path to (item i of vols)
				
				-- Gather basic info
				tell application "Finder" to set vol_name to name of disk of ((POSIX file vol_path) as alias)
				
				(* This is goofy, but for some reason Finder refuses to give me 
					a string for this if asked in the app, even though it works 
					when run via Script Editor directly. *)
				set vol_format to do shell script "osascript -e 'tell application \"Finder\" to return format of disk \"" & vol_name & "\"'"
				
				set thisReport to thisReport & {"OS version:", os_version, "", "Volume info:", "Name:" & tab & vol_name, "Path:" & tab & vol_path, "Format:" & tab & vol_format, "", "Current date:", currentdate}
				
				set hiddenfile to do shell script "find " & quoted form of vol_path & " -maxdepth 1 -name '.keepThisDriveSpinning'"
				
				if hiddenfile is not "" then
					set lastupdate to do shell script "cat " & quoted form of hiddenfile
					set thisReport to thisReport & {"", "Contents of hidden file:", lastupdate}
				end if
				
				-- Do we have a loaded launch agent for this volume?
				set agentstatus to ""
				try
					set launchagent to do shell script "grep -l " & quoted form of vol_path & " ~/Library/LaunchAgents/info.stovell.KeepDriveSpinning*"
				on error
					set launchagent to ""
				end try
				if launchagent is not "" then
					set launchagent_contents to do shell script "cat " & launchagent
					set label to do shell script "xmllint --xpath " & quoted form of "//string[preceding-sibling::key/text()='Label']/text()" & space & quoted form of launchagent
					set interval to do shell script "xmllint --xpath " & quoted form of "//integer[preceding-sibling::key/text()='StartInterval']/text()" & space & quoted form of launchagent
					set launchagent_status to do shell script "launchctl list | grep -i " & quoted form of (label & "$")
					set thisReport to thisReport & {"", "Launch agent path:", launchagent, "", "Launch agent status:", launchagent_status, "", "Launch agent contents:", launchagent_contents}
					set agentstatus to do shell script "launchctl list | grep -i " & quoted form of (label & "$") & " | cut -f 2 -"
				end if
				
				-- Okay, so we have a loaded launch agent. Is the hidden file writable? If not, is it locked?
				if hiddenfile is not "" then
					try
						do shell script "test -w " & quoted form of hiddenfile
						set w to true
						set l to false
					on error
						set w to false
						tell application "Finder" to set l to locked of ((POSIX file (hiddenfile)) as alias)
					end try
					
					if w then
						set thisReport to thisReport & {"", "Writing to the hidden file:"}
						try
							set writeattempt to do shell script "echo " & quoted form of lastupdate & " > " & quoted form of hiddenfile & " && echo Succeeded"
						on error errMsg
							set writeattempt to errMsg
						end try
						set thisReport to thisReport & {writeattempt}
					end if
				end if
				
				-- How long since the hidden file was last updated?
				set elapsed to missing value
				try
					set lastupdate_seconds to do shell script "date -j -f '%F %T %Z' '+%s' " & quoted form of lastupdate
					set elapsed to currentdate_seconds - lastupdate_seconds
				end try
				
				-- Output assessment
				if launchagent is "" or agentstatus is "" then
					set msg to {vol_name & " is not currently configured to be kept awake.", "Run Keep Drive Spinning to change this."}
					
				else if hiddenfile is "" then
					set hiddenfile to ".keepThisDriveSpinning"
					if last character of vol_path is not "/" then set hiddenfile to "/" & hiddenfile
					set hiddenfile to vol_path & hiddenfile
					set msg to {hiddenfile & " does not exist.", "This file must exist to keep " & vol_name & " awake.", "Run Keep Drive Spinning again to fix this."}
					
				else if l then
					set msg to {hiddenfile & " is locked.", "Run Keep Drive Spinning again to fix this."}
					tell application "Finder" to open information window of ((POSIX file hiddenfile) as alias)
					
				else if not w then
					set msg to {hiddenfile & " cannot be modified.", "This file must be modifiable to keep " & vol_name & " awake."}
					tell application "Finder" to open information window of ((POSIX file hiddenfile) as alias)
					
				else if elapsed is less than interval and agentstatus is "0" then
					set msg to {"The Keep Drive Spinning launch agent for " & vol_name & " is working correctly."}
					if interval is greater than 60 then
						set msg to msg & {"The drive is currently updated every " & interval & " seconds.", "If the drive is sometimes falling asleep, try using 60 seconds."}
					else
						set msg to msg & {"If the drive falls asleep despite this, it may be suffering from a bad connection or failing hardware."}
					end if
					
				else
					set msg to {"The Keep Drive Spinning launch agent for " & vol_name & " is not working correctly.", "The reason for the failure could not be determined automatically.", "The diagnostic information below may yield a clue."}
				end if
				
				if report is not {} then set report to report & {"", "==============================", ""}
				
				set report to report & {"Result:", "-------", msg, "", "", "Details:", "--------"} & thisReport
			on error errMsg
				set report to report & {"", "==============================", "", "An error occured:", errMsg}
			end try
		end repeat
		
		set report to concatenateList(return, report)
		ignoring application responses
			do shell script "now=$(date '+%F %H.%M.%S'); echo " & quoted form of report & " > ~/Desktop/\"Keep Drive Spinning diagnostic $now.txt\"; open -e ~/Desktop/\"Keep Drive Spinning diagnostic $now.txt\""
		end ignoring
		activate
	end runDiagnostic
	
	on uniqueList(stringList)
		set uniqueStrings to {}
		repeat with i from 1 to count stringList
			set str to (item i of stringList) as text
			if str is not "" and uniqueStrings does not contain str then
				set end of uniqueStrings to str
			end if
		end repeat
		return uniqueStrings
	end uniqueList
	
	on delimitText(delims, sourceText)
		set ASTID to AppleScript's text item delimiters
		set AppleScript's text item delimiters to delims
		set sourceText to text items of sourceText
		set AppleScript's text item delimiters to ASTID
		return sourceText
	end delimitText
	
	on findReplace(findText, replaceText, sourceText)
		set ASTID to AppleScript's text item delimiters
		set AppleScript's text item delimiters to findText
		set sourceText to text items of sourceText
		set AppleScript's text item delimiters to replaceText
		set sourceText to sourceText as text
		set AppleScript's text item delimiters to ASTID
		return sourceText
	end findReplace
	
	on concatenateList(delim, sourceText)
		set ASTID to AppleScript's text item delimiters
		set AppleScript's text item delimiters to delim
		set sourceText to sourceText as text
		set AppleScript's text item delimiters to ASTID
		return sourceText
	end concatenateList
	
	on applicationShouldTerminate:sender
		-- Insert code here to do any housekeeping before your application quits
		return current application's NSTerminateNow
	end applicationShouldTerminate:
	
end script
